#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <pwd.h>
#include <sys/types.h>
#include <dlfcn.h>

#include <mysql.h>

#define PW_NAME_MAX	64
#define PW_PASSWD_MAX	64
#define PW_GECOS_MAX	64
#define PW_DIR_MAX	64
#define PW_SHELL_MAX	64

static struct passwd *(*orig_getpwnam)(const char *);
static struct passwd *(*orig_getpwuid)(uid_t);

static MYSQL *my;

struct mysql_connect_params {
  char		*host;
  char		*user;
  char		*passwd;
  char		*db;
  unsigned int	port;
  char		*sock;
  unsigned int	flag;
} p;

char *query_getpwnam = "select name,passwd,uid,gid,gecos,dir,shell from passwd where name='%s'";
char *query_getpwuid = "select name,passwd,uid,gid,gecos,dir,shell from passwd where uid='%d'";

void puterr(const char *msg, const char *value)
{
  if (getenv("MYSQL_PW_DEBUG") || getenv("MYSQL_PW_ABORT"))
    fprintf(stderr, "%s: %s\n", msg, value);
}

void putdebug(const char *msg)
{
  if (getenv("MYSQL_PW_DEBUG"))
    fprintf(stderr, "%s\n", msg);
}

void set_params(struct mysql_connect_params *p)
{
  char *conf;
  FILE *fp;

  memset(p, 0, sizeof(*p));

  if ((conf = getenv("MYSQL_PW_CONF")) == NULL)
    conf = "/usr/local/etc/mysql_pw.conf";

  if ((fp = fopen(conf, "r")) != NULL) {
    char buf[1024];

    while (fgets(buf, sizeof(buf), fp)) {
      char *s;
      size_t n;

      if (buf[0] == '#')
	continue;
      if (buf[strlen(buf)-1] == '\n')
	buf[strlen(buf)-1] = '\0';
      if ((s = strchr(buf, ':')) == NULL)
	continue;
      *s++ = '\0';
      s += strspn(s, " \t");
      if (strcmp(buf, "host") == 0) {
	p->host = strdup(s);
      } else if (strcmp(buf, "user") == 0) {
	p->user = strdup(s);
      } else if (strcmp(buf, "passwd") == 0) {
	p->passwd = strdup(s);
      } else if (strcmp(buf, "db") == 0) {
	p->db = strdup(s);
      } else if (strcmp(buf, "port") == 0) {
	p->port = atoi(s);
      } else if (strcmp(buf, "sock") == 0) {
	p->sock = strdup(s);
      } else if (strcmp(buf, "flag") == 0) {
	p->flag = atoi(s);
      } else if (strcmp(buf, "query_getpwnam") == 0) {
	query_getpwnam = strdup(s);
      } else if (strcmp(buf, "query_getpwuid") == 0) {
	query_getpwuid = strdup(s);
      }
    }
  }
}


int try_connect(void)
{
  static int connection;

  if (! connection) {
    putdebug("pre mysql_real_connect");
    if (mysql_real_connect(my, p.host, p.user, p.passwd, p.db, p.port, p.sock, p.flag) == NULL) {
      puterr("try_connect", mysql_error(my));
      return 0;
    }
    connection = 1;
  }    
  putdebug("pre mysql_ping");
  if (mysql_ping(my) != 0) {
    puterr("try_connect", mysql_error(my));
    return 0;
  }
  return 1;
}

struct passwd *get_pw(char *query)
{
  static struct passwd pwd;
  static char pw_name[PW_NAME_MAX+1];
  static char pw_passwd[PW_PASSWD_MAX+1];
  static uid_t pw_uid;
  static gid_t pw_gid;
  static char pw_gecos[PW_GECOS_MAX+1];
  static char pw_dir[PW_DIR_MAX+1];
  static char pw_shell[PW_SHELL_MAX+1];

  MYSQL_RES *res = NULL;
  MYSQL_ROW row;

  putdebug("pre mysql_query");
  if (mysql_query(my, query) != 0) {
    puterr("get_pw: mysql_query", mysql_error(my));
    goto error;
  }
  putdebug("pre mysql_store_result");
  if ((res = mysql_store_result(my)) == NULL) {
    puterr("get_pw: mysql_store_result", mysql_error(my));
    goto error;
  }
  putdebug("pre mysql_num_rows");
  if (mysql_num_rows(res) == 0) {
    goto not_found;
  }
  putdebug("pre mysql_fetch_row");
  if ((row = mysql_fetch_row(res)) == NULL) {
    puterr("get_pw: mysql_fetch_row", mysql_error(my));
    goto error;
  }

  strcpy(pw_name, row[0]);
  strcpy(pw_passwd, row[1]);
  pw_uid = atoi(row[2]);
  pw_gid = atoi(row[3]);
  strcpy(pw_gecos, row[4]);
  strcpy(pw_dir, row[5]);
  strcpy(pw_shell, row[6]);

  pwd.pw_name = pw_name;
  pwd.pw_passwd = pw_passwd;
  pwd.pw_uid = pw_uid;
  pwd.pw_gid = pw_gid;
  pwd.pw_gecos = pw_gecos;
  pwd.pw_dir = pw_dir;
  pwd.pw_shell = pw_shell;

  mysql_free_result(res);
  return &pwd;

 error:
  if (res)
    mysql_free_result(res);
  return (struct passwd *)-1;

 not_found:
  if (res)
    mysql_free_result(res);
  return NULL;
}


void _init()
{
  void* dl;

  putdebug("enter _init");
  if ((dl = dlopen("libc.so.6", RTLD_LAZY)) == NULL) {
    puterr("_init", dlerror());
    exit(1);
  }
  if ((orig_getpwnam = dlsym(dl, "getpwnam")) == NULL) {
    puterr("_init", dlerror());
    exit(1);
  }
  if ((orig_getpwuid = dlsym(dl, "getpwuid")) == NULL) {
    puterr("_init", dlerror());
    exit(1);
  }

  set_params(&p);
  my = mysql_init(NULL);
}

struct passwd *getpwnam(const char *name)
{
  char name2[PW_NAME_MAX*2+1];
  char *sql;
  struct passwd *pwd;

  putdebug("pre orig_getpwnam");
  if ((pwd = orig_getpwnam(name)) != NULL)
    return pwd;
  putdebug("after orig_getpwnam");

  if (try_connect() == 0)
    goto error;
  if (strlen(name) > PW_NAME_MAX) {
    puterr("getpwnam: too large name", name);
    goto error;
  }
  putdebug("pre mysql_real_escape_string");
  mysql_real_escape_string(my, name2, name, strlen(name));
  putdebug("after mysql_real_escape_string");
  if ((sql = malloc(strlen(query_getpwnam)+strlen(name2))) == NULL) {
    puterr("getpwnam", "malloc failed");
    goto error;
  }
  sprintf(sql, query_getpwnam, name2);
  if ((pwd = get_pw(sql)) != (struct passwd *)-1) {
    free(sql);
    return pwd;
  }
  free(sql);

 error:
  if (getenv("MYSQL_PW_ABORT")) {
    fprintf(stderr, "mysql_pw module: getpwnam:  abort\n");
    exit(1);
  }
  return NULL;
}


struct passwd *getpwuid(uid_t uid)
{
  char *sql;
  struct passwd *pwd;

  if ((pwd = orig_getpwuid(uid)) != NULL)
    return pwd;
  if (try_connect() == 0)
    goto error;
  if ((sql = malloc(strlen(query_getpwuid)+20)) == NULL) {
    puterr("getpwuid", "malloc failed");
    goto error;
  }
  sprintf(sql, query_getpwuid, uid);
  if ((pwd = get_pw(sql)) != (struct passwd *)-1) {
    free(sql);
    return pwd;
  }
  free(sql);

error:
  if (getenv("MYSQL_PW_ABORT")) {
    fprintf(stderr, "mysql_pw module: getpwuid: abort\n");
    exit(1);
  }
  return NULL;
}


void _fini()
{
  mysql_close(my);
}
